<?php

namespace Dreamscape;

use Dreamscape\Response\Base;
use Illuminate\Support\Arr;
use Illuminate\Support\Facades\Cache;
use Illuminate\Support\Facades\Log;

use Dreamscape\Response\Generic;

final class API
{
	private const LOGKEY = 'API';
	private const CACHE_TIME = 60*60*12;
	private const MAX_LIMIT = 100;

	// Our reseller details
	private string $key;
	private string $url;

	public function __construct(string $id,string $key,bool $tryprod=FALSE)
	{
		$this->url = (config('app.env') == 'local' && ! $tryprod) ? 'https://reseller-api.sandbox.ds.network' : 'https://reseller-api.ds.network';
		$this->key = $key;

		Log::debug(sprintf('%s:Dreamscape API for id [%s]',static::LOGKEY,$id));
	}

	/**
	 * Call the API
	 *
	 * @param string $path
	 * @param array $parameters
	 * @return object|array
	 * @throws \Exception
	 */
	private function execute(string $path,array $parameters=[])
	{
		$url = sprintf('%s/%s',$this->url,$path);
		$request_id = md5(uniqid().microtime(true));
		$signature = md5($request_id.$this->key);
		$method = Arr::get($parameters,'method','GET');

		if ($parameters)
			Arr::forget($parameters,'method');

		// If we are passed an array, we'll do a normal post.
		switch ($method) {
			case 'GET':
				$request = $this->prepareRequest(
					$url,
					$parameters,
					[
						'accept: application/json',
						'Api-Request-Id: '.$request_id,
						'Api-Signature: '.$signature,
					]
				);
				break;

			case 'POST':
				$request = $this->prepareRequestPost(
					$url,
					$parameters,
					[
						'accept: application/json',
						'Api-Request-Id: '.$request_id,
						'Api-Signature: '.$signature,
					]
				);
				break;

			case 'PUT':
				$request = $this->prepareRequestPut(
					$url,
					$parameters,
					[
						'accept: application/json',
						'Api-Request-Id: '.$request_id,
						'Api-Signature: '.$signature,
					]
				);
				break;

			default:
				throw new \Exception(sprintf('Unknown method: %s',$method));
		}

		$key = md5($path.serialize($parameters));

		//Cache::forget($key);
		$result = Cache::remember($key,self::CACHE_TIME,function() use ($request,$url) {
			try {
				$response = curl_exec($request);

				switch($x=curl_getinfo($request,CURLINFO_HTTP_CODE)) {
					case 400:
					case 401:
					case 404:
						dump([$xx=curl_getinfo($request),'response'=>$response]);

						throw new \Exception(sprintf('CURL exec returned %d: %s (%s)',$x,curl_error($request),serialize($xx)));
				}

				curl_close($request);
				return json_decode($response);

			} catch (\Exception $e) {
				dump(['error'=>$e->getMessage()]);
				Log::error(sprintf('%s:Got an error while posting to [%s] (%s)',static::LOGKEY,$url,$e->getMessage()),['m'=>__METHOD__]);

				curl_close($request);
				throw new \Exception($e->getMessage());
			}
		});

		return $result;
	}

	/**
	 * Get a list of our clients
	 *
	 * @param array $parameters
	 * @return Generic
	 * @throws \Exception
	 */
	public function getCustomers(array $parameters=[]): Generic
	{
		Log::debug(sprintf('%s:Get a list of customers',static::LOGKEY));
		$key = 'customers';

		$fetchall = FALSE;
		if (Arr::get($parameters,'fetchall')) {
			Arr::forget($parameters,['fetchall']);
			$parameters['limit'] = self::MAX_LIMIT;
			$parameters['page'] = 1;

			$fetchall = TRUE;
		}

		$o = new Generic($this->execute($key,$parameters),$key);

		return $fetchall ? $this->fetchall($o,$key,$parameters) : $o;
	}

	private function fetchall(Base $o,string $path,array $parameters): Base
	{
		while ($o->hasMore()) {
			$parameters['page']++;

			$o->merge($this->execute($path,$parameters));
		}

		return $o;
	}

	/**
	 * Get a list of our clients domains
	 *
	 * @param array $parameters
	 * @return Generic
	 * @throws \Exception
	 */
	public function getDomains(array $parameters=[]): Generic
	{
		Log::debug(sprintf('%s:Get a list of domains',static::LOGKEY));

		return new Generic($this->execute('domains',$parameters),'domains');
	}

	/**
	 * Get a list of our available product types
	 *
	 * @return Generic
	 * @throws \Exception
	 */
	public function getProductTypes(): Generic
	{
		Log::debug(sprintf('%s:Get a list of product types',static::LOGKEY));

		return new Generic($this->execute('products/types'));
	}

	/**
	 * Setup the API call
	 *
	 * @param        $url
	 * @param array $parameters
	 * @param array $headers
	 * @return resource
	 */
	private function prepareRequest($url,array $parameters=[],array $headers=[])
	{
		$request = curl_init();

		curl_setopt($request,CURLOPT_URL,$url.($parameters ? '?'.http_build_query($parameters) : ''));
		curl_setopt($request,CURLOPT_RETURNTRANSFER,TRUE);
		if ($headers)
			curl_setopt($request,CURLOPT_HTTPHEADER,$headers);
		curl_setopt($request,CURLINFO_HEADER_OUT,TRUE);
		curl_setopt($request,CURLOPT_SSL_VERIFYPEER,FALSE);

		return $request;
	}

	private function prepareRequestPost($url,$parameters='',$headers=[])
	{
		$request = $this->prepareRequest($url,[],$headers);
		curl_setopt($request,CURLOPT_POST,TRUE);
		curl_setopt($request,CURLOPT_POSTFIELDS,json_encode($parameters));

		return $request;
	}

	private function prepareRequestPut($url,$parameters='',$headers=[])
	{
		$request = $this->prepareRequest($url,[],$headers);
		//curl_setopt($request,CURLOPT_PUT,TRUE);
		curl_setopt($request,CURLOPT_CUSTOMREQUEST,'PUT');
		curl_setopt($request,CURLOPT_POSTFIELDS,json_encode($parameters));

		return $request;
	}
}