<?php

namespace Dreamscape\Response;

use Illuminate\Support\Collection;
use Illuminate\Support\Facades\App;
use Illuminate\Support\Facades\Log;

use Dreamscape\Models\{Customer,Domain};

/**
 * This parent class handles responses received from Trello
 *
 * @note: This class is used for events not specifically created.
 */
abstract class Base implements \JsonSerializable, \Countable, \ArrayAccess, \Iterator
{
	protected const LOGKEY = 'RB-';

	protected Collection $_data;
	protected bool $_status;
	protected object $_pagination;
	protected ?string $_type;

	private ?int $counter = NULL;
	protected const TYPES = [
		'customers',
		'domains',
	];

	/**
	 * Default Constructor Setup
	 *
	 * @param object $response
	 * @param string $type
	 * @throws \Exception
	 */
	public function __construct(object $response,string $type)
	{
		if (! in_array($type,self::TYPES))
			throw new \Exception('Unknown data type: '.$type);

		$this->_data = $this->data($response,$type);
		$this->_status = $response->status;
		$this->_pagination = $response->pagination;
		$this->_type = $type;

		// This is only for child classes
		if (get_class($this) == Base::class) {
			Log::debug(sprintf('%s:Dreamscape RESPONSE Initialised [%s]',static::LOGKEY,get_class($this)),['m'=>__METHOD__]);

			if (App::environment() == 'dev')
				file_put_contents('/tmp/response',print_r($this,TRUE),FILE_APPEND);
		}
	}

	/* ABSTRACT */

	/**
	 * When we json_encode this object, this is the data that will be returned
	 */
	public function jsonSerialize(): mixed
	{
		return $this->_data ?: new \stdClass;
	}

	public function current(): mixed
	{
		return $this->_data[$this->counter];
	}

	public function next(): void
	{
		$this->counter++;
	}

	public function key(): mixed
	{
		return $this->counter;
	}

	public function valid(): bool
	{
		return isset($this->_data[$this->counter]);
	}

	public function rewind(): void
	{
		$this->counter = 0;
	}

	public function offsetExists(mixed $offset): bool
	{
		return $this->has($offset);
	}

	public function offsetGet(mixed $offset): mixed
	{
		return $this->_data->get($offset);
	}

	public function offsetSet(mixed $offset, mixed $value): void
	{
		throw new \Exception('Method not implemented: '.__METHOD__);
	}

	public function offsetUnset(mixed $offset): void
	{
		$this->_data->forget($offset);

		// Rekey the collection
		$this->_data = $this->_data->values();

		// Reset the counter if we have deleted a value before it
		if ($offset < $this->counter)
			$this->counter--;
	}

	public function count(): int
	{
		return count($this->_data);
	}

	/* METHODS */

	/**
	 * Convert our response into a collection of the appropriate model
	 *
	 * @param object $response
	 * @param string $type
	 * @return Collection
	 * @throws \Exception
	 */
	private function data(object $response,string $type): Collection
	{
		switch ($type) {
			case 'customers':
				$data = collect(Customer::hydrate($response->data));
				break;

			case 'domains':
				$data = collect(Domain::hydrate($response->data));
				break;

			default: throw new \Exception('Unknown object type: '.$this->_type);
		}

		return $data;
	}

	/**
	 * Determine if there are more values to obtain
	 *
	 * @return bool
	 */
	public function hasMore(): bool
	{
		return $this->_status && ($this->count() < $this->totalItems());
	}

	/**
	 * Merge more values to the result
	 *
	 * @param object $response
	 * @return void
	 * @throws \Exception
	 */
	public function merge(object $response): void
	{
		$this->_data = $this->_data->merge($this->data($response,$this->_type));
	}

	/**
	 * Return the current resultset page number
	 *
	 * @return int
	 */
	public function page(): int
	{
		return $this->_pagination->current_page;
	}

	/**
	 * Search for an item in the result
	 *
	 * @param string $key
	 * @param mixed $value
	 * @return mixed
	 */
	public function search(string $key, mixed $value): mixed
	{
		return $this->_data->search(function($item) use ($key,$value) { return $item->{$key} == $value; });
	}

	/**
	 * Return the result status
	 *
	 * @return bool
	 */
	public function status(): bool
	{
		return $this->_status;
	}

	/**
	 * Return the total items available
	 *
	 * @return int
	 */
	public function totalItems(): int
	{
		return $this->_pagination->total_items;
	}
}